#pragma once

#include <stdbool.h>
#include <stddef.h>
#include "config.h"

bool spawn(const char *cwd, char *const argv[],
           int stdin_fd, int stdout_fd, int stderr_fd);

bool spawn_expand_template(
    const struct config_spawn_template *template,
    size_t key_count, const char *key_names[static key_count],
    const char *key_values[static key_count], size_t *argc, char ***argv);
